package ru.landgrafhomyak.bgtu.networks0.icmp

import kotlinx.atomicfu.AtomicRef
import kotlinx.atomicfu.atomic
import kotlinx.coroutines.TimeoutCancellationException
import kotlinx.coroutines.withTimeout
import kotlinx.datetime.Clock
import kotlinx.datetime.Instant
import ru.landgrafhomyak.bgtu.networks0.low_level.sockets.IcmpSocket
import ru.landgrafhomyak.bgtu.networks0.utilities.compareAndExchange
import ru.landgrafhomyak.utility.IntSerializers
import ru.landgrafhomyak.utility.highlevel_try_finally.safeAutoClose1
import ru.landgrafhomyak.utility.highlevel_try_finally.safeAutoClose2


@OptIn(ExperimentalUnsignedTypes::class)
class Pinger1 : AutoCloseable {
	private enum class State {
		READY, CONNECTED, CLOSED
	}

	private val _rawSocket: IcmpSocket
	private var _state: AtomicRef<State>


	internal constructor(rawSocket: IcmpSocket) {
		safeAutoClose2(onError = rawSocket::close) {
			this._state = atomic(State.READY)
			this._rawSocket = rawSocket
		}
	}

	private class ShortKeyPair(val identifier: UShort, val seqNo: UShort) {
		override fun equals(other: Any?): Boolean {
			if (other !is ShortKeyPair) return false
			return this.identifier == other.identifier && this.seqNo == other.seqNo
		}

		override fun hashCode(): Int {
			return ((this.identifier.toUInt() shl 16) or (this.seqNo).toUInt()).toInt()
		}

		override fun toString(): String = "<icmp ping key pair identifier=${this.identifier} seq_no=${this.seqNo}>"
	}

	private fun _generateKeyPair(): ShortKeyPair {
		val timestamp = Clock.System.now()
		return ShortKeyPair(
			identifier = (timestamp.epochSeconds ushr 13).toULong().toUShort(),
			seqNo = ((timestamp.epochSeconds shl 3) or ((((timestamp.nanosecondsOfSecond ushr 27) and 7).toLong())) and 0xFFFF).toULong().toUShort(),
		)
	}

	suspend fun ping(timeoutMillis: UInt): UInt? {
		when (this._state.compareAndExchange(State.READY, State.CONNECTED)) {
			State.CLOSED -> throw IllegalStateException("Pinger is closed")
			State.CONNECTED -> throw IllegalStateException("Pinger is busy with another thread/coroutine")
			State.READY -> {}
		}
		safeAutoClose1(finally = { this._state.value = State.READY }) {
			val start: Instant
			val key = this._generateKeyPair()
			val header = IcmpPingHeader(identifier_be = key.identifier, seqNo_be = key.seqNo)
			val serialized = header.serialize(Pinger1.payloadLikeInPingProgramm)
			start = Clock.System.now()
			this._rawSocket.sendRaw(serialized)
			try {
				withTimeout(timeoutMillis.toULong().toLong()) {
					while (true)
						if (this@Pinger1._dispatchResponse(key, this@Pinger1._rawSocket.recvRaw())) return@withTimeout
				}
			} catch (_: TimeoutCancellationException) {
				return null
			}

			val td = Clock.System.now() - start
			return td.inWholeMilliseconds.toUInt()
		}
	}

	private fun _dispatchResponse(expected: ShortKeyPair, raw: UByteArray): Boolean {
		if (raw.size < 8) return false
		val actualKey = ShortKeyPair(identifier = IntSerializers.decode16beHu(raw, 4), seqNo = IntSerializers.decode16beHu(raw, 6))
		val echoHeader = IcmpEchoHeader(identifier_be = actualKey.identifier, seqNo_be = actualKey.seqNo)
		if (actualKey != expected) return false
		if (IntSerializers.decode8Bu(raw, 0) != echoHeader.type) return false
		if (IntSerializers.decode8Bu(raw, 1) != echoHeader.code) return false
		val data = raw.sliceArray(8..<(raw.size))
		if (!data.contentEquals(Pinger1.payloadLikeInPingProgramm)) return false
		if (echoHeader.checksum(data) != IntSerializers.decode16beHu(raw, 2)) return false
		return true
	}

	override fun close() {
		when (this._state.compareAndExchange(State.READY, State.CLOSED)) {
			State.CLOSED -> throw IllegalStateException("Pinger is closed")
			State.CONNECTED -> throw IllegalStateException("Can't close pinger while it measures time")
			State.READY -> {}
		}
		this._rawSocket.close()
	}

	companion object {
		@Suppress("SpellCheckingInspection")
		private val payloadLikeInPingProgramm = "abcdefghijklmnopqrstuvwabcdefghi".encodeToByteArray().asUByteArray()
	}
}