package ru.landgrafhomyak.bgtu.networks0.low_level.sockets

import kotlin.coroutines.Continuation
import kotlin.coroutines.resume
import kotlinx.cinterop.CValue
import kotlinx.cinterop.ExperimentalForeignApi
import kotlinx.cinterop.StableRef
import kotlinx.cinterop.allocArray
import kotlinx.cinterop.asStableRef
import kotlinx.cinterop.cValue
import kotlinx.cinterop.get
import kotlinx.cinterop.memScoped
import kotlinx.cinterop.ptr
import kotlinx.coroutines.suspendCancellableCoroutine
import platform.linux.EPOLLIN
import platform.linux.EPOLLONESHOT
import platform.linux.EPOLLOUT
import platform.linux.EPOLL_CTL_ADD
import platform.linux.EPOLL_CTL_DEL
import platform.linux.EPOLL_CTL_MOD
import platform.linux.epoll_create
import platform.linux.epoll_ctl
import platform.linux.epoll_event
import platform.linux.epoll_wait
import platform.posix.close as close_lowlevel
import platform.linux.inet_pton
import platform.posix.AF_INET
import platform.posix.sockaddr_in
import platform.posix.sockaddr_in6
import ru.landgrafhomyak.bgtu.networks0.low_level.c_interop_utilities.PosixUtilities
import ru.landgrafhomyak.bgtu.networks0.low_level.multithreading.withLock
import ru.landgrafhomyak.utility.highlevel_try_finally.safeAutoClose1
import ru.landgrafhomyak.utility.highlevel_try_finally.safeAutoClose2
import ru.landgrafhomyak.utility.highlevel_try_finally.tryFinallyChain
import ru.landrafhomyak.utility.reference_counter.CloseableReferenceCounter
import ru.landgrafhomyak.bgtu.networks0.low_level.multithreading.Mutex as TMutex

@OptIn(ExperimentalForeignApi::class)
@Suppress("JoinDeclarationAndAssignment", "ConvertSecondaryConstructorToPrimary", "FunctionName")
class EpollSocketEventLoop : SocketEventLoopScope.Closeable {
	private val _epollDescriptor: Int
	private val _socketsCount: CloseableReferenceCounter

	constructor() {
		this._socketsCount = CloseableReferenceCounter("Epoll event-loop is closed")
		this._epollDescriptor = epoll_create(1)
		if (this._epollDescriptor < 0)
			PosixUtilities.throwErrno { d -> RuntimeException("Failed to create epoll descriptor: $d") }
	}

	private fun _parseIpv4(rawAddr: String, port: UShort): CValue<sockaddr_in> = cValue<sockaddr_in> out@{
		memScoped {
			val err = inet_pton(AF_INET, rawAddr, this@out.sin_addr.ptr)
			when {
				err == 0 -> throw IllegalArgumentException("IPv4 address is ill-formatted")
				err < 0 -> PosixUtilities.throwErrno { d -> throw RuntimeException("Failed to parse IPv4 address: $d") }
			}
		}
		this@out.sin_family = AF_INET.toUShort()
		this@out.sin_port = port
	}

	override fun icmp_IPv4(addr: String): IcmpSocket {
		return BoundIcmpSocket(this._parseIpv4(addr, 0u))
	}

	override fun icmp_IPv6(addr: String): IcmpSocket {
		TODO("Not yet implemented")
	}

	private fun _schedule(metadata: _SocketMetadata) {
		val ee = cValue<epoll_event> {
			this.events = ((if (metadata.read == null) 0 else EPOLLIN) or (if (metadata.write == null) 0 else EPOLLOUT) or EPOLLONESHOT).toUInt()
			this.data.fd = metadata.fd
			this.data.ptr = metadata.stableRef.asCPointer()
		}
		if (metadata.read != null || metadata.write != null) {
			if (!metadata.isAddedToEpoll) {
				if (0 != epoll_ctl(this._epollDescriptor, EPOLL_CTL_ADD, metadata.fd, ee))
					PosixUtilities.throwErrno { d -> RuntimeException("Failed to add fd to epoll: $d") }
				metadata.isAddedToEpoll = true
			} else {
				if (0 != epoll_ctl(this._epollDescriptor, EPOLL_CTL_MOD, metadata.fd, ee))
					PosixUtilities.throwErrno { d -> RuntimeException("Failed to modify fd in epoll: $d") }
			}
		} else {
			if (metadata.isAddedToEpoll) {
				if (0 != epoll_ctl(this._epollDescriptor, EPOLL_CTL_DEL, metadata.fd, null))
					PosixUtilities.throwErrno { d -> RuntimeException("Failed to remove fd from epoll: $d") }
				metadata.isAddedToEpoll = false
			}
		}
	}

	private fun _remove(metadata: _SocketMetadata) {
		metadata.sync.withLock {
			if (metadata.isAddedToEpoll) {
				if (0 != epoll_ctl(this._epollDescriptor, EPOLL_CTL_DEL, metadata.fd, null))
					PosixUtilities.throwErrno { d -> RuntimeException("Failed to remove fd from epoll: $d") }
			}
		}
	}

	private suspend fun _waitForRead(metadata: _SocketMetadata) {
		suspendCancellableCoroutine { continuation ->
			metadata.sync.withLock {
				if (metadata.read != null)
					throw IllegalStateException("Socket already waiting for read")
				metadata.read = continuation
				this._schedule(metadata)
				continuation.invokeOnCancellation {
					metadata.sync.withLock {
						metadata.read = null
						this._schedule(metadata)
					}
				}
			}
		}
	}


	private suspend fun _waitForWrite(metadata: _SocketMetadata) {
		suspendCancellableCoroutine { continuation ->
			metadata.sync.withLock {
				if (metadata.write != null)
					throw IllegalStateException("Socket already waiting for write")
				metadata.write = continuation
				this._schedule(metadata)
				continuation.invokeOnCancellation {
					metadata.sync.withLock {
						metadata.write = null
						this._schedule(metadata)
					}
				}
			}
		}
	}

	private fun _resume(ev: epoll_event) {
		val metadata = (ev.data.ptr ?: throw RuntimeException("Lost reference to socket metadata")).asStableRef<_SocketMetadata>().get()
		metadata.sync.withLock {
			if ((ev.events and EPOLLIN.toUInt()) != 0u) {
				(metadata.read ?: throw RuntimeException("Socket ready for read, but nobody requests")).resume(Unit)
				metadata.read = null
			}
			if ((ev.events and EPOLLOUT.toUInt()) != 0u) {
				(metadata.write ?: throw RuntimeException("Socket ready for write, but nobody requests")).resume(Unit)
				metadata.write = null
			}
			if (metadata.read != null || metadata.write != null)
				this._schedule(metadata)
		}
	}

	fun runUntilError(eventsBufferSize: Int = 1000) {
		memScoped {
			val readyEvents = allocArray<epoll_event>(eventsBufferSize)
			while (true) {
				val readyEventsCount = epoll_wait(this@EpollSocketEventLoop._epollDescriptor, readyEvents, eventsBufferSize, -1)
				if (readyEventsCount < 0)
					PosixUtilities.throwErrno { d -> RuntimeException("Error while waiting epoll: $d") }

				for (i in 0..<readyEventsCount) {
					this@EpollSocketEventLoop._resume(readyEvents[i])
				}
			}
		}
	}


	override fun close() {
		this._socketsCount.close("There are still opened sockets bound to this event-loop")
		if (0 != close_lowlevel(this._epollDescriptor))
			PosixUtilities.throwErrno { d -> RuntimeException("Failed to close epoll descriptor: $d") }
	}

	class _SocketMetadata : AutoCloseable {
		val fd: Int
		var read: Continuation<Unit>?
		var write: Continuation<Unit>?
		val sync: TMutex
		val stableRef: StableRef<_SocketMetadata>
		var isAddedToEpoll: Boolean

		constructor(fd: Int) {
			this.isAddedToEpoll = false
			this.fd = fd
			this.read = null
			this.write = null
			this.stableRef = StableRef.create(this)
			safeAutoClose2(onError = this.stableRef::dispose) {
				this.sync = TMutex()
			}
		}

		override fun close() {
			safeAutoClose1(
				action = this.sync::close,
				finally = this.stableRef::dispose
			)
		}
	}


	private inner class BoundIcmpSocket : NonblockingIcmpSocketImpl {
		private val __metadata: _SocketMetadata

		constructor(ipv4: CValue<sockaddr_in>) : super(ipv4) {
			safeAutoClose2(onError = { super.close() }) {
				this@EpollSocketEventLoop._socketsCount.tryIncref {
					this.__metadata = _SocketMetadata(this._socketFd)
				}
			}
		}

		constructor(ipv6: CValue<sockaddr_in6>) : super(ipv6) {
			safeAutoClose2(onError = { super.close() }) {
				this@EpollSocketEventLoop._socketsCount.tryIncref {
					this.__metadata = _SocketMetadata(this._socketFd)
				}
			}
		}


		override suspend fun _waitForWrite() =
			this@EpollSocketEventLoop._waitForWrite(this.__metadata)

		override suspend fun _waitForRead() =
			this@EpollSocketEventLoop._waitForRead(this.__metadata)

		override fun _free() {
			tryFinallyChain { chain ->
				chain.action { this@EpollSocketEventLoop._remove(this.__metadata) }
				chain.action { this@EpollSocketEventLoop._socketsCount.decref() }
				chain.action { this.__metadata.close() }
				chain.action { super._free() }
			}
		}
	}
}
